/* z8dis.c */

/* A z8 (8681/82) Disassembler for PC's .

  Author: Bob Bush (Compuserve-73105,1332) 08-Jul-1990

  Copyright (C) 1990 R.Bush. All rights reserved.

  This program is freely distributable as long as this copyright notice
  is retained. It intended for personal, non-commercial use.

  Compiled with MicroSoft 'C' Ver.5.1, Compact memory model
  
*/
   
#include <string.h>
#include <stdio.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <ctype.h>
#include <malloc.h>
#include <fcntl.h>
#include <memory.h>

#define S_IREAD     0000400         /* read permission, owner */
#define S_IWRITE    0000200         /* write permission, owner */


char *get_cc();


/* define the different operand formations */
/* note: 0 is invalid.. (illegal opcode) */

#define 	R1	1
#define		IR1	2
#define		r1r2	3
#define		r1Ir2	4
#define		R2R1	5
#define		IR2R1	6
#define		R1IM	7
#define		r1R2	8
#define		r2R1	9
#define		ccRA	10
#define		r1IM	11
#define		ccDA	12
#define		r1	13
#define		IR1IM	14
#define		r1RA	15
#define		Implied 16
#define		IRR1	17
#define		IM	18
#define		RR1	19
#define		r1Irr2	20
#define		Ir1Irr2 21
#define		r2Irr1  22
#define		Ir2Irr1 23
#define		r1xR2   24
#define		r2xR1	25
#define		DA	26
#define		Ir1r2	27
#define		R2IR1	28

/* condition codes. note: hi nibble is the condition code */

struct cc_code {
    char *mne;
    int  cc_value;
};

struct cc_code cc[] = {
    "c",0x70,
    "nc",0xf0,
    "z",0x60,
    "nz",0xe0,
    "pl",0xd0,
    "mi",0x50,
    "ov",0x40,
    "nov",0xc0,
    "eq",0x60,
    "ne",0xe0,
    "ge",0x90,
    "lt",0x10,
    "gt",0xa0,
    "le",0x20,
    "uge",0xf0,
    "ult",0x70,
    "ugt",0xb0,
    "ule",0x30,
    "always",0x80,
    "never",0x00
};

#define CCSIZE sizeof (cc) / sizeof(cc[0])

struct opcode {
        char    *mne;
	int	operands;
        int     bytecount;
};

/* lookup table of opcodes */

struct opcode optab[] = {
	"dec    ",R1,		2,	/* 00 */
	"dec	",IR1,		2,	/* 01 */
	"add    ",r1r2,		2,	/* 02 */
	"add    ",r1Ir2,	2,	/* 03 */
	"add    ",R2R1,		3,	/* 04 */
	"add    ",IR2R1,	3,	/* 05 */
	"add    ",R1IM,		3,	/* 06 */
	"add    ",IR1IM,	3,	/* 07 */
	"ld     ",r1R2,		2,	/* 08 */
	"ld     ",r2R1,		2,	/* 09 */
	"djnz   ",r1RA,		2,	/* 0a */
	"jr     ",ccRA,		2,	/* 0b */
	"ld     ",r1IM,		2,	/* 0c */
	"jp     ",ccDA,		3,	/* 0d */
	"inc    ",r1,		1,	/* 0e */
	"invalid",0,		1,	/* 0f */

	"rlc    ",R1,		2,	/* 10 */
	"rlc    ",IR1,		2,	/* 11 */
	"adc    ",r1r2,		2,	/* 12 */
	"adc    ",r1Ir2,	2,	/* 13 */
	"adc    ",R2R1,		3,	/* 14 */
	"adc    ",IR2R1,	3,	/* 15 */
	"adc    ",R1IM,		3,	/* 16 */
	"adc    ",IR1IM,	3,	/* 17 */
	"ld     ",r1R2,		2,	/* 18 */
	"ld     ",r2R1,		2,	/* 19 */
	"djnz   ",r1RA,		2,	/* 1a */
	"jr     ",ccRA,		2,	/* 1b */
	"ld     ",r1IM,		2,	/* 1c */
	"jp     ",ccDA,		3,	/* 1d */
	"inc    ",r1,		1,	/* 1e */
	"invalid",0,		1,	/* 1f */

	"inc    ",R1,		2,	/* 20 */
	"inc    ",IR1,		2,	/* 21 */
	"sub    ",r1r2, 	2,	/* 22 */
	"sub    ",r1Ir2,	2,	/* 23 */
	"sub    ",R2R1, 	3,	/* 24 */
	"sub    ",IR2R1,	3,	/* 25 */
	"sub    ",R1IM,		3,	/* 26 */
	"sub    ",IR1IM,	3,	/* 27 */
	"ld     ",r1R2,		2,	/* 28 */
	"ld     ",r2R1,		2,	/* 29 */
	"djnz   ",r1RA,		2,	/* 2a */
	"jr     ",ccRA,		2,	/* 2b */
	"ld     ",r1IM,		2,	/* 2c */
	"jp     ",ccDA,		3,	/* 2d */
	"inc    ",r1,		1,	/* 2e */
	"invalid",0,		1,	/* 2f */

	"jp     ",IRR1,		2,	/* 30 */
	"srp    ",IM,		2,	/* 31 */
	"sbc    ",r1r2, 	2,	/* 32 */
	"sbc    ",r1Ir2,	2,	/* 33 */
	"sbc    ",R2R1, 	3,	/* 34 */
	"sbc    ",IR2R1,	3,	/* 35 */
	"sbc    ",R1IM,		3,	/* 36 */
	"sbc    ",IR1IM,	3,	/* 37 */
	"ld     ",r1R2,		2,	/* 38 */
	"ld     ",r2R1,		2,	/* 39 */
	"djnz   ",r1RA,		2,	/* 3a */
	"jr     ",ccRA,		2,	/* 3b */
	"ld     ",r1IM,		2,	/* 3c */
	"jp     ",ccDA,		3,	/* 3d */
	"inc    ",r1,		1,	/* 3e */
	"invalid",0,		1,	/* 3f */

	"da     ",R1,		2,	/* 40 */
	"da     ",IR1,		2,	/* 41 */
	"or     ",r1r2, 	2,	/* 42 */
	"or     ",r1Ir2,	2,	/* 43 */
	"or     ",R2R1, 	3,	/* 44 */
	"or     ",IR2R1,	3,	/* 45 */
	"or     ",R1IM,		3,	/* 46 */
	"or     ",IR1IM,	3,	/* 47 */
	"ld     ",r1R2,		2,	/* 48 */
	"ld     ",r2R1,		2,	/* 49 */
	"djnz   ",r1RA,		2,	/* 4a */
	"jr     ",ccRA,		2,	/* 4b */
	"ld     ",r1IM,		2,	/* 4c */
	"jp     ",ccDA,		3,	/* 4d */
	"inc    ",r1,		1,	/* 4e */
	"invalid",0,		1,	/* 4f */

	"pop    ",R1,		2,	/* 50 */
	"pop    ",IR1,		2,	/* 51 */
	"and    ",r1r2, 	2,	/* 52 */
	"and    ",r1Ir2,	2,	/* 53 */
	"and    ",R2R1, 	3,	/* 54 */
	"and    ",IR2R1,	3,	/* 55 */
	"and    ",R1IM, 	3,	/* 56 */
	"and    ",IR1IM,	3,	/* 57 */
	"ld     ",r1R2,		2,	/* 58 */
	"ld     ",r2R1,		2,	/* 59 */
	"djnz   ",r1RA,		2,	/* 5a */
	"jr     ",ccRA,		2,	/* 5b */
	"ld     ",r1IM,		2,	/* 5c */
	"jp     ",ccDA,		3,	/* 5d */
	"inc    ",r1,		1,	/* 5e */
	"invalid",0,		1,	/* 5f */

	"com    ",R1,		2,	/* 60 */
	"com    ",IR1,		2,	/* 61 */
	"tcm    ",r1r2, 	2,	/* 62 */
	"tcm    ",r1Ir2,	2,	/* 63 */
	"tcm    ",R2R1, 	3,	/* 64 */
	"tcm    ",IR2R1,	3,	/* 65 */
	"tcm    ",R1IM,		3,	/* 66 */
	"tcm    ",IR1IM,	3,	/* 67 */
	"ld     ",r1R2,		2,	/* 68 */
	"ld     ",r2R1,		2,	/* 69 */
	"djnz   ",r1RA,		2,	/* 6a */
	"jr     ",ccRA,		2,	/* 6b */
	"ld     ",r1IM,		2,	/* 6c */
	"jp     ",ccDA,		3,	/* 6d */
	"inc    ",r1,		1,	/* 6e */
	"invalid",0,		1,	/* 6f */

	"push   ",R1,		2,	/* 70  evaluates same as R1 */
	"push   ",IR1,		2,	/* 71  evaluates same as IR1 */
	"tm     ",r1r2, 	2,	/* 72 */
	"tm     ",r1Ir2,	2,	/* 73 */
	"tm     ",R2R1, 	3,	/* 74 */
	"tm     ",IR2R1,	3,	/* 75 */
	"tm     ",R1IM,		3,	/* 76 */
	"tm     ",IR1IM,	3,	/* 77 */
	"ld     ",r1R2,		2,	/* 78 */
	"ld     ",r2R1,		2,	/* 79 */
	"djnz   ",r1RA,		2,	/* 7a */
	"jr     ",ccRA,		2,	/* 7b */
	"ld     ",r1IM,		2,	/* 7c */
	"jp     ",ccDA,		3,	/* 7d */
	"inc    ",r1,		1,	/* 7e */
	"invalid",0,		1,	/* 7f */

	"decw   ",RR1,		2,	/* 80 */
	"decw	",IR1,		2,	/* 81 */
	"lde    ",r1Irr2,	2,	/* 82 */
	"ldei   ",Ir1Irr2,	2,	/* 83 */
	"invalid",0,		1,	/* 84 */
	"invalid",0,		1,	/* 85 */
	"invalid",0,		1,	/* 86 */
	"invalid",0,		1,	/* 87 */
	"ld     ",r1R2,		2,	/* 88 */
	"ld     ",r2R1,		2,	/* 89 */
	"djnz   ",r1RA,		2,	/* 8a */
	"jr     ",ccRA,		2,	/* 8b */
	"ld     ",r1IM,		2,	/* 8c */
	"jp     ",ccDA,		3,	/* 8d */
	"inc    ",r1,		1,	/* 8e */
	"di     ",Implied,	1,	/* 8f */

	"rl     ",R1,		2,	/* 90 */
	"rl     ",IR1,		2,	/* 91 */
	"lde    ",r2Irr1,	2,	/* 92 */
	"ldei   ",Ir2Irr1,	2,	/* 93 */
	"invalid",0,		1,	/* 94 */
	"invalid",0,		1,	/* 95 */
	"invalid",0,		1,	/* 96 */
	"invalid",0,		1,	/* 97 */
	"ld     ",r1R2,		2,	/* 98 */
	"ld     ",r2R1,		2,	/* 99 */
	"djnz   ",r1RA,		2,	/* 9a */
	"jr     ",ccRA,		2,	/* 9b */
	"ld     ",r1IM,		2,	/* 9c */
	"jp     ",ccDA,		3,	/* 9d */
	"inc    ",r1,		1,	/* 9e */
	"ei     ",Implied,	1,	/* 9f */

	"incw   ",RR1,		2,	/* a0 */
	"incw	",IR1,		2,	/* a1 */
	"cp     ",r1r2, 	2,	/* a2 */
	"cp     ",r1Ir2,	2,	/* a3 */
	"cp     ",R2R1, 	3,	/* a4 */
	"cp     ",IR2R1,	3,	/* a5 */
	"cp     ",R1IM,		3,	/* a6 */
	"cp     ",IR1IM,	3,	/* a7 */
	"ld     ",r1R2,		2,	/* a8 */
	"ld     ",r2R1,		2,	/* a9 */
	"djnz   ",r1RA,		2,	/* aa */
	"jr     ",ccRA,		2,	/* ab */
	"ld     ",r1IM,		2,	/* ac */
	"jp     ",ccDA,		3,	/* ad */
	"inc    ",r1,		1,	/* ae */
	"ret    ",Implied,	1,	/* af */

	"clr    ",R1,		2,	/* b0 */
	"clr    ",IR1,		2,	/* b1 */
	"xor    ",r1r2, 	2,	/* b2 */
	"xor    ",r1Ir2,	2,	/* b3 */
	"xor    ",R2R1, 	3,	/* b4 */
	"xor    ",IR2R1,	3,	/* b5 */
	"xor    ",R1IM,		3,	/* b6 */
	"xor    ",IR1IM,	3,	/* b7 */
	"ld     ",r1R2,		2,	/* b8 */
	"ld     ",r2R1,		2,	/* b9 */
	"djnz   ",r1RA,		2,	/* ba */
	"jr     ",ccRA,		2,	/* bb */
	"ld     ",r1IM,		2,	/* bc */
	"jp     ",ccDA,		3,	/* bd */
	"inc    ",r1,		1,	/* be */
	"iret   ",Implied,	1,	/* bf */

	"rrc    ",R1,		2,	/* c0 */
	"rrc    ",IR1,		2,	/* c1 */
	"ldc    ",r1Irr2,	2,	/* c2 */
	"ldci   ",Ir1Irr2,	2,	/* c3 */
	"invalid",0,		1,	/* c4 */
	"invalid",0,		1,	/* c5 */
	"invalid",0,		1,	/* c6 */
	"ld     ",r1xR2,	3,	/* c7 */
	"ld     ",r1R2,		2,	/* c8 */
	"ld     ",r2R1,		2,	/* c9 */
	"djnz   ",r1RA,		2,	/* ca */
	"jr     ",ccRA,		2,	/* cb */
	"ld     ",r1IM,		2,	/* cc */
	"jp     ",ccDA,		3,	/* cd */
	"inc    ",r1,		1,	/* ce */
	"rcf    ",Implied,	1,	/* cf */

	"sra    ",R1,		2,	/* d0 */
	"sra    ",IR1,		2,	/* d1 */
	"ldc    ",r2Irr1,	2,	/* d2 */
	"ldci   ",Ir2Irr1,	2,	/* d3 */
	"call   ",IRR1,		2,	/* d4 */
	"invalid",0,		1,	/* d5 */
	"call   ",DA,		3,	/* d6 */
	"ld     ",r2xR1,	3,	/* d7 */
	"ld     ",r1R2,		2,	/* d8 */
	"ld     ",r2R1,		2,	/* d9 */
	"djnz   ",r1RA,		2,	/* da */
	"jr     ",ccRA,		2,	/* db */
	"ld     ",r1IM,		2,	/* dc */
	"jp     ",ccDA,		3,	/* dd */
	"inc    ",r1,		1,	/* de */
	"scf    ",Implied,	1,	/* df */

	"rr     ",R1,		2,	/* e0 */
	"rr     ",IR1,		2,	/* e1 */
	"invalid",0,		1,	/* e2 */
	"ld     ",r1Ir2,	2,	/* e3 */
	"ld     ",R2R1, 	3,	/* e4 */
	"ld     ",IR2R1,	3,	/* e5 */
	"ld     ",R1IM,		3,	/* e6 */
	"ld     ",IR1IM,	3,	/* e7 */
	"ld     ",r1R2,		2,	/* e8 */
	"ld     ",r2R1,		2,	/* e9 */
	"djnz   ",r1RA,		2,	/* ea */
	"jr     ",ccRA,		2,	/* eb */
	"ld     ",r1IM,		2,	/* ec */
	"jp     ",ccDA,		3,	/* ed */
	"inc    ",r1,		1,	/* ee */
	"ccf    ",Implied,	1,	/* ef */

	"swap   ",R1,		2,	/* f0 */
	"swap   ",IR1,		2,	/* f1 */
	"invalid",0,		1,	/* f2 */
	"ld     ",Ir1r2,	2,	/* f3 */
	"invalid",0,		1,	/* f4 */
	"ld     ",R2IR1,	3,      /* f5 */
	"invalid",0,		1,	/* f6 */
	"invalid",0,		1,	/* f7 */
	"ld     ",r1R2,		2,	/* f8 */
	"ld     ",r2R1,		2,	/* f9 */
	"djnz   ",r1RA,		2,	/* fa */
	"jr     ",ccRA,		2,	/* fb */
	"ld     ",r1IM,		2,	/* fc */
	"jp     ",ccDA,		3,	/* fd */
	"inc    ",r1,		1,	/* fe */
	"nop   ",Implied,	1,	/* ff */
};

/* Command token buffer */

char *cmd[] = {
    "                                          ",
    "                                          ",
    "                                          ",
    "                                          ",
};

unsigned char huge *buffer;
unsigned char huge *bin_pc,*buf_pc;
unsigned char huge *bufbase;

unsigned int pc,hex_start;
FILE *fpin = NULL;
FILE *fpout = NULL;
int outfile;
char buff[80]; /* input string buffer */
char infname[80];
char outfname[80];

int file_flag = 0;
int hex_listing  = 0;
int uf_flag = 0;
int radix;

main()
{
    int cmd;

    printf("\nZ8 Disassembler Ver 1.0\n\n");
    if(!(buffer = halloc(65536L,sizeof(char))))
	pgm_exit("Couldn't allocate required memory");

    bufbase = buffer;pc = 0;
    menu();
    while((cmd = get_cmd()) != 99);
    pgm_exit("End Program");
}

menu()
{
    printf("rb <filename> (addr) Read binary file into buffer starting at addr\n");
    printf("ri <filename> Read Intel File into buffer\n");
    printf("rm <filename> Read Motorola File into buffer\n");
    printf("u  (pcvalue)  disasm @ pcvalue for 20 lines\n");
    printf("uf (pcvalue)  disasm @ pcvalue for 20 lines, also output to file\n");
    printf("d  (hexval)   dumpmem @ hexval for 256 bytes\n");
    printf("a  (hexval)   Alter memory starting at hexval\n");
    printf("bf <filename> (hexval) Read Basic textfile into mem @ hexval\n");
    printf("wb <filename> <start addr> <end addr> Write binary file from buffer\n");
    printf("?  Display this menu\n");
    printf("x  Exit program\n");
}

/* get a command, break apart tokens, process command */

get_cmd()
{
    int tkindex;
    char *token;

    tkindex = 0;
    printf("\nEnter Command:");
    gets(buff);
    token = strtok(buff,", ");
    while(token != NULL)
	{
	    strncpy(cmd[tkindex],token,40);
	    tkindex++;
	    token = strtok(NULL,", ");
	}
     if(tkindex) {
	if(strcmp(cmd[0],"rb") == 0) {
	   if(tkindex > 2)
	       read_bin(cmd[1],getval(cmd[2]));
	   else {
	       read_bin(cmd[1],0);
	       printf("\nDefault load addr: 0000\n");
	   }
        } 
	else
	if(strcmp(cmd[0],"ri") == 0) {
	   read_intel(cmd[1]);
	}
	else
	if(strcmp(cmd[0],"rm") == 0) {
	   read_motorola(cmd[1]);
	}
	else
	if(strcmp(cmd[0],"wb") == 0) {
	   if(tkindex < 4)
		printf("\nMissing required parameters\n");
	   else
	   	write_bin(cmd[1],getval(cmd[2]),getval(cmd[3]));
	}
	else
	if(strcmp(cmd[0],"a") == 0) {
	   if(tkindex > 1) 
	       set_pc(cmd[1]);
	   alter_mem();
	}
	else
	if(strcmp(cmd[0],"d") == 0) {
	   if(tkindex > 1) {
	       pc = getval(cmd[1]);
	       pc &= 0xff00;		/* limit to page */
	       buf_pc = bufbase+pc;	/* add in buffer offset */
	    /*   set_pc(cmd[1]); */
	   }
	   dmpmem();
	}
        else
	if(strcmp(cmd[0],"x") == 0)
	   return(99);
	else
	if(strcmp(cmd[0],"?") == 0)
	   menu();
	else
	if(strcmp(cmd[0],"u") == 0) {
	   if(tkindex > 1) 
	       set_pc(cmd[1]);
	       disasm();
	}
	else
	if(strcmp(cmd[0],"uf") == 0) {
	   if(tkindex > 1) 
	       set_pc(cmd[1]);
	   if(!fpout)
		setup_outfile();
	   uf_flag = 1;
	   disasm();
	}
	else
	if(strcmp(cmd[0],"bf") == 0) {
	   if(tkindex > 2)
	   	read_basic(cmd[1],getval(cmd[2]));
	   else {
		read_basic(cmd[1],0x1020);
		printf("\nLoaded at Default Autostart Addr: 0x1020\n");
		}
	}

     }
     return(1);
}

setup_outfile()
{
    printf("\nEnter a filename for disasm output:");
    gets(outfname);
    if(!(fpout = fopen(outfname,"w")))
	printf("Can't open output file\n");
    else {
	file_flag = 1;
	printf("Do you want hex listing included (y/n):");
	gets(buff);
	if(buff[0] == 'y' || buff[0] == 'Y')
	   hex_listing = 1;
        else
	   hex_listing = 0;
    }
}

		
write_bin(fname,start,end)
char *fname;
int start,end;
{
    int binfile;
    int len;

    unlink(fname);
   
    if((binfile = open(fname,O_CREAT|O_BINARY|O_RDWR,S_IWRITE)) == -1) {
        printf("\n Unable to open output file\n");
	return(0);
    }
    len = (end-start) + 1;
    if(len == 0) {
	len = 65535;
        write(binfile,buffer+start,65535); /* max for one write */
        write(binfile,buffer+65535L,1);
    }
    else
    write(binfile,buffer+start,len);

    close(binfile);
}

alter_mem()
{
    int val;

    printf("\nAlter memory -- Enter '.' to exit\n");

    for(;;) {
	printf("%04x: %02x ",pc, *buf_pc & 0xff);
	gets(buff);
	if(buff[0] == '.')
	    break;
	if(buff[0] != '\0') {
	    val = getval(buff);
            *buf_pc = (unsigned char)val;
	}
	inc_pcs(1);
    }
}


/* char digit connversion routines */

cvdig(ch)			/* return value of digit or -1 */
char ch;
{
 int z;
	if((z = ch) >= '0' && z <= '9') z -= '0';
           else
        	if((z &= 0137) >= 'A' && z <= 'F') z -= 'A' - 10;
           else
        	return (-1);
	if(z < radix) {
	   return(z);
	}
  return (-1);
}

/* get hex numeric value from string */

getval(curline)
char *curline;
{
  int j,n,k;

  radix = 16;
  j = 0;k = 0;
  while(!(isxdigit(*curline)))
	curline++;

  if(radix < 255) {
      while(*curline) {
	    if((n = cvdig(*curline++)) < 0){
	       curline--;
	       break;}
	    j = j * radix + n;
	    k++;
      }
  }
  else {
	j = *curline++;
	curline++;
  }
  return(j);
}

/* the Meat of the disassembler */

disasm()
{
    int lc;
    unsigned int index;
    unsigned int save_pc;
    unsigned char huge *save_buf_pc;
    
    save_pc = pc; save_buf_pc = buf_pc;

    for(lc = 0; lc < 20; lc++) {
	printf("%04x  ",pc);
	index = *buf_pc;
	disp_hex(buf_pc,optab[index].bytecount);
	printf("%s",optab[index].mne);
	display_operands(index,buf_pc);
	printf("\n");
	inc_pcs(optab[index].bytecount);
     }
    if(file_flag && uf_flag) {
	pc = save_pc; buf_pc = save_buf_pc;   
	fdisasm();
    }
}

disp_hex(buf,count)
char huge *buf;
int count;
{
    switch(count) {
	case 1:	printf("%02x        ",*buf & 0xff);
		break;
	case 2: printf("%02x %02x     ",*buf	& 0xff,*(buf+1) & 0xff);
		break;
	case 3: printf("%02x %02x %02x  ",*buf & 0xff,*(buf+1) & 0xff,*(buf+2) & 0xff);
		break;
	}
}

display_operands(index,bufptr)
int index;
char huge *bufptr;
{
    switch(optab[index].operands) {
	case R1:	bufptr++;
			if(working(bufptr))
			   printf("r%d",*bufptr & 0x0f);
			else
			   printf("r:%d",*bufptr & 0xff);
			break;
	case IR1:	bufptr++;
			if(working(bufptr))
			   printf("@r%d",*bufptr & 0x0f);
			else
			   printf("@r:%d",*bufptr & 0xff);
			break;
	case r1r2:	bufptr++;
			printf("r%d,r%d",*bufptr >> 4 & 0x0f,*bufptr & 0x0f);
			break;
	case r1Ir2:	bufptr++;
			printf("r%d,@r%d",*bufptr >> 4 & 0x0f,*bufptr & 0x0f);
			break;
	case R2R1:	bufptr+=2;
			if(working(bufptr))
			   printf("r%d,",*bufptr & 0x0f);
			else
			   printf("r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   printf("r%d",*bufptr & 0x0f);
			else
			   printf("r:%d",*bufptr & 0xff);
			break;
	case IR2R1:	bufptr+=2;
			if(working(bufptr))
			   printf("@r%d,",*bufptr & 0x0f);
			else
			   printf("@r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   printf("r%d",*bufptr & 0x0f);
			else
			   printf("r:%d",*bufptr & 0xff);
			break;
	case R1IM:	bufptr++;
			if(working(bufptr))
			   printf("r%d,",*bufptr & 0x0f);
			else
			   printf("r:%d,",*bufptr & 0xff);
			bufptr++;
			printf("#%d",*bufptr & 0xff);
			break;
	case r1R2:	printf("r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			if(working(bufptr))
			   printf("r%d",*bufptr & 0x0f);
			else
			   printf("r:%d",*bufptr & 0xff);
			break;	
	case r2R1:	bufptr++;
			if(working(bufptr))
			   printf("r%d,",*bufptr & 0x0f);
			else
			   printf("r:%d,",*bufptr & 0xff);
			bufptr--;
			printf("r%d",*bufptr >> 4 & 0x0f);
			break;
	case r1RA:	printf("r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			printf("$%04x",(pc+2) + *bufptr);
			break;	
	case ccRA:	printf("%s,",get_cc(*bufptr & 0xf0));
			bufptr++;
			printf("$%04x",(pc+2) + *bufptr);
			break;
	case r1IM:	printf("r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			printf("#%d",*bufptr & 0xff);
			break;
	case r1:	printf("r%d",*bufptr >> 4 & 0x0f);
			break;
	case IRR1:	bufptr++;
			if(working(bufptr))
			   printf("@rr%d",*bufptr & 0x0f);
			else
			   printf("@rr:%d",*bufptr & 0xff);
			break;
	case IM:	bufptr++;
			printf("#%d",*bufptr & 0xff);
			break;
	case RR1:	bufptr++;
			if(working(bufptr))
			   printf("rr%d",*bufptr & 0x0f);
			else
			   printf("rr:%d",*bufptr & 0xff);
			break;
	case r1Irr2:	bufptr++;
			printf("r%d,",*bufptr >> 4 & 0x0f);
			printf("@rr%d",*bufptr & 0x0f);
			break;
	case Ir1Irr2:	bufptr++;
			printf("@r%d,",*bufptr >> 4 & 0x0f);
			printf("@rr%d",*bufptr & 0x0f);
			break;
	case r2Irr1:	bufptr++;
			printf("@rr%d,",*bufptr & 0x0f);
			printf("r%d",*bufptr >> 4 & 0x0f);
			break;
	case Ir2Irr1:   bufptr++;
			printf("@rr%d,",*bufptr & 0x0f);
			printf("@r%d",*bufptr >> 4 & 0x0f);
			break;
	case r1xR2:	bufptr++;
			printf("r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			printf("r:%d",*bufptr & 0xff);
			bufptr--;
			printf("(r%d)",*bufptr & 0x0f);
			break;
	case r2xR1:	bufptr+=2;
			printf("r:%d",*bufptr & 0xff);
			bufptr--;
			printf("(r%d),",*bufptr & 0x0f);
			printf("r%d",*bufptr >> 4 & 0x0f);
			break;
	case ccDA:	printf("%s,",get_cc(*bufptr & 0xf0));
			bufptr++;
			printf("$%02x",*bufptr & 0xff);
			bufptr++;
			printf("%02x",*bufptr & 0xff);
			break;
	case DA:	bufptr++;
			printf("$%02x",*bufptr & 0xff);
			bufptr++;
			printf("%02x",*bufptr & 0xff);
			break;
	case Ir1r2:	bufptr++;
			printf("@r%d,",*bufptr >> 4 & 0x0f);
			printf("r%d",*bufptr & 0x0f);
			break;
	case R2IR1:	bufptr+=2;
			if(working(bufptr))
			   printf("@r%d,",*bufptr & 0x0f);
			else
			   printf("@r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   printf("r%d",*bufptr & 0x0f);
			else
			   printf("r:%d",*bufptr & 0xff);
			break;
	case 0:		printf("???"); /* illegal */
			break;
	case Implied:	break;
    }
}


fdisasm()
{
    int lc;
    unsigned int index;

    uf_flag = 0;    
    for(lc = 0; lc < 20; lc++) {
	if(hex_listing)
	    fprintf(fpout,"%04x  ",pc);
	index = *buf_pc;
	if(hex_listing)
	     fdisp_hex(buf_pc,optab[index].bytecount);
	if(hex_listing)
	    fprintf(fpout,"%s",optab[index].mne);
	else
	    fprintf(fpout,"        %s",optab[index].mne);
	fdisplay_operands(index,buf_pc);
	fprintf(fpout,"\n");
	inc_pcs(optab[index].bytecount);
     }
}

fdisp_hex(buf,count)
char *buf;
int count;
{
    switch(count) {
	case 1:	fprintf(fpout,"%02x        ",*buf & 0xff);
		break;
	case 2: fprintf(fpout,"%02x %02x     ",*buf	& 0xff,*(buf+1) & 0xff);
		break;
	case 3: fprintf(fpout,"%02x %02x %02x  ",*buf & 0xff,*(buf+1) & 0xff,*(buf+2) & 0xff);
		break;
	}
}

fdisplay_operands(index,bufptr)
int index;
char *bufptr;
{
    switch(optab[index].operands) {
	case R1:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d",*bufptr & 0xff);
			break;
	case IR1:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"@r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"@r:%d",*bufptr & 0xff);
			break;
	case r1r2:	bufptr++;
			fprintf(fpout,"r%d,r%d",*bufptr >> 4 & 0x0f,*bufptr & 0x0f);
			break;
	case r1Ir2:	bufptr++;
			fprintf(fpout,"r%d,@r%d",*bufptr >> 4 & 0x0f,*bufptr & 0x0f);
			break;
	case R2R1:	bufptr+=2;
			if(working(bufptr))
			   fprintf(fpout,"r%d,",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   fprintf(fpout,"r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d",*bufptr & 0xff);
			break;
	case IR2R1:	bufptr+=2;
			if(working(bufptr))
			   fprintf(fpout,"@r%d,",*bufptr & 0x0f);
			else
			   fprintf(fpout,"@r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   fprintf(fpout,"r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d",*bufptr & 0xff);
			break;
	case R1IM:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"r%d,",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d,",*bufptr & 0xff);
			bufptr++;
			fprintf(fpout,"#%d",*bufptr & 0xff);
			break;
	case r1R2:	fprintf(fpout,"r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d",*bufptr & 0xff);
			break;	
	case r2R1:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"r%d,",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d,",*bufptr & 0xff);
			bufptr--;
			fprintf(fpout,"r%d",*bufptr >> 4 & 0x0f);
			break;
	case r1RA:	fprintf(fpout,"r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			fprintf(fpout,"$%04x",(pc+2) + *bufptr);
			break;	
	case ccRA:	fprintf(fpout,"%s,",get_cc(*bufptr & 0xf0));
			bufptr++;
			fprintf(fpout,"$%04x",(pc+2) + *bufptr);
			break;
	case r1IM:	fprintf(fpout,"r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			fprintf(fpout,"#%d",*bufptr & 0xff);
			break;
	case r1:	fprintf(fpout,"r%d",*bufptr >> 4 & 0x0f);
			break;
	case IRR1:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"@rr%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"@rr:%d",*bufptr & 0xff);
			break;
	case IM:	bufptr++;
			fprintf(fpout,"#%d",*bufptr & 0xff);
			break;
	case RR1:	bufptr++;
			if(working(bufptr))
			   fprintf(fpout,"rr%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"rr:%d",*bufptr & 0xff);
			break;
	case r1Irr2:	bufptr++;
			fprintf(fpout,"r%d,",*bufptr >> 4 & 0x0f);
			fprintf(fpout,"@rr%d",*bufptr & 0x0f);
			break;
	case Ir1Irr2:	bufptr++;
			fprintf(fpout,"@r%d,",*bufptr >> 4 & 0x0f);
			fprintf(fpout,"@rr%d",*bufptr & 0x0f);
			break;
	case r2Irr1:	bufptr++;
			fprintf(fpout,"@rr%d,",*bufptr & 0x0f);
			fprintf(fpout,"r%d",*bufptr >> 4 & 0x0f);
			break;
	case Ir2Irr1:   bufptr++;
			fprintf(fpout,"@rr%d,",*bufptr & 0x0f);
			fprintf(fpout,"@r%d",*bufptr >> 4 & 0x0f);
			break;
	case r1xR2:	bufptr++;
			fprintf(fpout,"r%d,",*bufptr >> 4 & 0x0f);
			bufptr++;
			fprintf(fpout,"r:%d",*bufptr & 0xff);
			bufptr--;
			fprintf(fpout,"(r%d)",*bufptr & 0x0f);
			break;
	case r2xR1:	bufptr+=2;
			fprintf(fpout,"r:%d",*bufptr & 0xff);
			bufptr--;
			fprintf(fpout,"(r%d),",*bufptr & 0x0f);
			fprintf(fpout,"r%d",*bufptr >> 4 & 0x0f);
			break;
	case ccDA:	fprintf(fpout,"%s,",get_cc(*bufptr & 0xf0));
			bufptr++;
			fprintf(fpout,"$%02x",*bufptr & 0xff);
			bufptr++;
			fprintf(fpout,"%02x",*bufptr & 0xff);
			break;
	case DA:	bufptr++;
			fprintf(fpout,"$%02x",*bufptr & 0xff);
			bufptr++;
			fprintf(fpout,"%02x",*bufptr & 0xff);
			break;
	case Ir1r2:	bufptr++;
			fprintf(fpout,"@r%d,",*bufptr >> 4 & 0x0f);
			fprintf(fpout,"r%d",*bufptr & 0x0f);
			break;
	case R2IR1:	bufptr+=2;
			if(working(bufptr))
			   fprintf(fpout,"@r%d,",*bufptr & 0x0f);
			else
			   fprintf(fpout,"@r:%d,",*bufptr & 0xff);
			bufptr--;
			if(working(bufptr))
			   fprintf(fpout,"r%d",*bufptr & 0x0f);
			else
			   fprintf(fpout,"r:%d",*bufptr & 0xff);
			break;
	case 0:		fprintf(fpout,"???"); /* illegal */
			break;
	case Implied:	break;
    }
}


static char *badcc = "???";

char *get_cc(cond)
int cond;
{
    int j;


    for(j = 0; j < CCSIZE; j++) {
	if(cond == cc[j].cc_value)
	   return(cc[j].mne);
    }
    return(badcc);
}

working(ptr)
char *ptr;
{
    if((*ptr & 0xf0) == 0xe0)	/* pattern signifies working register */
	return(1);
    else
	return(0);
}

				

set_pc(s)
char *s;
{
    pc = getval(s);
    buf_pc = bufbase+pc;	/* add in buffer offset */
}

inc_pcs(val)
int val;
{
    pc += val;
    buf_pc += val;
}

read_basic(fname,offset)
char *fname;
unsigned int offset;
{
    FILE *basfile;
    int linenum;
    unsigned char *membuff;
    char lbuff[102];
    char slnum[10];
    char sltext[100];
    char *p;

    if(!(basfile = fopen(fname,"r"))){
	printf("Can't open input file\n");
	return(0);
    }
    membuff = (char*)buffer+offset;

    while ((p = fgets(lbuff,100,basfile)) != NULL) {
	get_bline(lbuff,slnum,sltext);
	linenum = atoi(slnum);
	*membuff++ = linenum >> 8 & 0xff;
        *membuff++ = linenum & 0xff;
        strcpy(membuff,sltext);
	membuff+=(strlen(sltext)+1);
    }
    *membuff++ = 0xff;
    *membuff = 0xff;

    fclose(basfile);
}
	
/* decompose line into linenum and text strings */

get_bline(s,num,txt)
char *s,*num,*txt;
{
  /* remove newline */

    while(*s != ' ') 
	*num++ = *s++;
    *num = '\0';

    s++;

    while(*s != '\n')
	*txt++ = *s++;
    *txt = '\0';
}

struct stat fil_stat;

read_bin(fname,addr)
char *fname;
unsigned int addr;
{
    int infile;
    long size;

    if((infile = open(fname,O_RDONLY|O_BINARY)) == -1) {
        printf("\n Unable to open input file\n");
	return(0);
    }

    fstat(infile,&fil_stat);
    size = fil_stat.st_size;

    if(size + (long)addr > 65536L) {
	close(infile);
	printf("\nError: File too big -or- filesize+offset would overwite buffer\n");
	return(0);
    }
    if(size == 65536L) {
        read(infile,buffer,65535); /* max for one read */
        read(infile,buffer+65535L,1);
    }
    else
        read(infile,buffer+addr,65535);
    printf("%ld bytes read\n",size);
    close(infile);
}


read_motorola(fname)
char *fname;
{
    int status;

    if(!(fpin = fopen(fname,"r"))){
	printf("Can't open input file\n");
	return(0);
    }
    printf("Loading Motorola Hex file\n");
    if(!(status = load_hex()))
	printf("Unable to load file (not Motorola format?)\n");
    else
	printf("\nHex load starting at %04x\n",hex_start);
    fclose(fpin);
    fpin = NULL;
}

read_intel(fname)
char *fname;
{
    int status;

    if(!(fpin = fopen(fname,"r"))) {
	printf("Can't open input file\n");
	return(0);
    }
    printf("Loading Intel Hex file\n");
    if(!(status = load_hexi())) 
	printf("Unable to load file (not Intel format?)\n");
    else
	printf("\nHex load starting at %04x\n",hex_start);
    fclose(fpin);
    fpin = NULL;
}

pgm_exit(s)
char *s;
{
    printf("\n%s\n",s);
    if(fpin)
	fclose(fpin);
    if(fpout)
	fclose(fpout);
    if(buffer)
	free(buffer);
    exit(0);
}

load_hex()
{
  int c,j,len,byte;
  int flag;

    flag = 1;

    while ((c = getc(fpin)) != EOF) {
	if (c == 'S') {
	    c=getc(fpin);
		if (c=='1') {
		    len = getbyte(fpin);
		    len -= 3;
		    bin_pc = buffer;
		    if(flag) {
			hex_start = getadr(fpin);
			flag = 0;
			bin_pc += hex_start;
		    }
		    else
		    bin_pc += getadr(fpin);
		    for(j=0;j<len;j++) {
			byte=getbyte(fpin);
			*bin_pc++ = byte;
		    }
		    c=getbyte(fpin);c=getc(fpin);
		}
	    else
		return(1);
	}
	else {
	   printf("Not motorola format\n");
	   return(0); }
     }
     return(1);
}

load_hexi()
{
  int c,j,len,byte;
  int flag;

    flag = 1;

    while ((c = getc(fpin)) != EOF) {
	if (c == ':') {
	    len=getbyte(fpin);
		if (len > 0) {
		    bin_pc = buffer;
		    if(flag) {
			hex_start = getadr(fpin);
			flag = 0;
			bin_pc += hex_start;
		    }
		    else
		        bin_pc += getadr(fpin);
		    getbyte(fpin);		/* zero byte */
		    for(j=0;j<len;j++) {
			byte=getbyte(fpin);
			*bin_pc++ = byte;
		    }
		    c=getbyte(fpin);c=getc(fpin);
		}
	    else
		return(1);
	}
	else {
	   printf("Not Intel format\n");
	   return(0); }
     }
     return(1);
}

getadr()
{
    static unsigned addrhi,lo;

    addrhi = getbyte(fpin);
    addrhi <<= 8;
    lo = getbyte(fpin);
    return(addrhi + lo);
}

getbyte()
{
    static unsigned lo,hi;

    hi = getc(fpin);
    lo = getc(fpin);

    hi -= (hi < 'A') ? 48 : 55;			/* ascii hex byte to binary */
    hi <<= 4;					/* shift to hi nibble */
    lo -= (lo < 'A') ? 48 : 55;

    return(hi + lo);				/* return binary */
}

/* dump 256 bytes in hex+ascii from pointer pm */

dmpmem()
{
   static int disp_pc;
   static char temp[100];
   static char huge *mp;
   static int sline,hline,aline,sb,sb1;
   static char *header[2] = {
      "\n       0  1  2  3  4  5  6  7  8  9",
      "  A  B  C  D  E  F   0123456789ABCDEF\n",
      };
   disp_pc = pc;
   mp = buf_pc;
   printf("%s%s",header[0],header[1]);
   for(sline = 0;sline < 16;sline++) {
      sb = 0;
      sb1 = sprintf(&temp[0],"\n%04x  ",disp_pc);
      sb += 7;
      for(hline = 0;hline < 16;hline++) {
            sb1 = sprintf(&temp[sb],"%02x ",(0xff & *mp));
            sb += 3;
            mp++;disp_pc++;
      }
            mp -= 16;
            sb1 = sprintf(&temp[sb],"  ");
            sb += 2;
            for(aline = 0; aline < 16;aline++) {
               if(*mp > 31 && *mp < 128)
                  temp[sb] = *mp;
               else
                  temp[sb] = '.';
               mp++;sb++;
            }
   temp[sb] = '\0';
   printf("%s",temp);
   }
   inc_pcs(256);
   printf("\n");
}
